@@ -1,41 +1,30 @@
-from __future__ import unicode_literals, division, absolute_import, print_function
-
+try:
+    from collections.abc import Callable # Python 3.6
+except ImportError as e:
+    from collections import Callable
 import sys
-PY3 = sys.version_info[0] == 3
-if PY3:
-    text_type = str
-    binary_type = bytes
-    unicode = str
-    basestring = str
-else:
-    range = xrange
-    text_type = unicode
-    binary_type = str
-    chr = unichr
+from collections import OrderedDict
 
 from pdb import set_trace
-import collections
 import re
 import warnings
 from sigil_bs4.dammit import EntitySubstitution
 
 DEFAULT_OUTPUT_ENCODING = "utf-8"
-PY3K = (sys.version_info[0] > 2)
 
 whitespace_re = re.compile("\s+")
 
-
 NON_BREAKING_INLINE_TAGS = ("a","abbr","acronym","b","bdo","big","br",
     "button","cite","code","del","dfn","em","font","i","image","img",
-    "input","ins","kbd","label","map","nobr","object","q","s","samp",
-    "select","small","span","strike","strong","sub","sup","textarea",
-    "tt","u","var","wbr","mbp:nu")
+    "input","ins","kbd","label","map","mark", "nobr","object","q",
+    "ruby","rt","s","samp","select","small","span","strike","strong",
+    "sub","sup","textarea","tt","u","var","wbr","mbp:nu")
 
-PRESERVE_WHITESPACE_TAGS = ("pre","textarea","script","style")
+PRESERVE_WHITESPACE_TAGS = ("code","pre","textarea","script","style")
 
 VOID_TAGS = ("area","base","basefont","bgsound","br","col","command",
-    "embed","event-source","frame","hr","image","img","input","keygen",
-    "link","menuitem","meta","param","source","spacer","track","wbr",
+    "embed","event-source","frame","hr","img","input","keygen",
+    "link","meta","param","source","spacer","track","wbr",
     "mbp:pagebreak")
 
 NO_ENTITY_SUB_TAGS = ("script", "style")
@@ -65,23 +54,23 @@ def _alias(attr):
     return alias
 
 
-class NamespacedAttribute(unicode):
+class NamespacedAttribute(str):
 
     def __new__(cls, prefix, name, namespace=None):
         if name is None:
-            obj = unicode.__new__(cls, prefix)
+            obj = str.__new__(cls, prefix)
 
         elif prefix is None:
             # Not really namespaced.
-            obj = unicode.__new__(cls, name)
+            obj = str.__new__(cls, name)
         else:
-            obj = unicode.__new__(cls, prefix + ":" + name)
+            obj = str.__new__(cls, prefix + ":" + name)
         obj.prefix = prefix
         obj.name = name
         obj.namespace = namespace
         return obj
 
-class AttributeValueWithCharsetSubstitution(unicode):
+class AttributeValueWithCharsetSubstitution(str):
     """A stand-in object for a character encoding specified in HTML."""
 
 class CharsetMetaAttributeValue(AttributeValueWithCharsetSubstitution):
@@ -92,7 +81,7 @@ class CharsetMetaAttributeValue(AttributeValueWithCharsetSubstitution):
     """
 
     def __new__(cls, original_value):
-        obj = unicode.__new__(cls, original_value)
+        obj = str.__new__(cls, original_value)
         obj.original_value = original_value
         return obj
 
@@ -115,9 +104,9 @@ class ContentMetaAttributeValue(AttributeValueWithCharsetSubstitution):
         match = cls.CHARSET_RE.search(original_value)
         if match is None:
             # No substitution necessary.
-            return unicode.__new__(str, original_value)
+            return str.__new__(str, original_value)
 
-        obj = unicode.__new__(cls, original_value)
+        obj = str.__new__(cls, original_value)
         obj.original_value = original_value
         return obj
 
@@ -197,7 +186,7 @@ class PageElement(object):
 
     def format_string(self, s, formatter='minimal'):
         """Format the given string using the given formatter."""
-        if not isinstance(formatter, collections.Callable):
+        if not isinstance(formatter, Callable):
             formatter = self._formatter_for_name(formatter)
         if formatter is None:
             output = s
@@ -340,7 +329,7 @@ class PageElement(object):
     def insert(self, position, new_child):
         if new_child is self:
             raise ValueError("Cannot insert a tag into itself.")
-        if (isinstance(new_child, basestring)
+        if (isinstance(new_child, str)
             and not isinstance(new_child, NavigableString)):
             new_child = NavigableString(new_child)
 
@@ -443,13 +432,13 @@ class PageElement(object):
         index = parent.index(self)
         parent.insert(index+1, successor)
 
-    def find_next(self, name=None, attrs={}, text=None, **kwargs):
+    def find_next(self, name=None, attrs=OrderedDict(), text=None, **kwargs):
         """Returns the first item that matches the given criteria and
         appears after this Tag in the document."""
         return self._find_one(self.find_all_next, name, attrs, text, **kwargs)
     findNext = find_next  # BS3
 
-    def find_all_next(self, name=None, attrs={}, text=None, limit=None,
+    def find_all_next(self, name=None, attrs=OrderedDict(), text=None, limit=None,
                     **kwargs):
         """Returns all items that match the given criteria and appear
         after this Tag in the document."""
@@ -457,14 +446,14 @@ class PageElement(object):
                              **kwargs)
     findAllNext = find_all_next  # BS3
 
-    def find_next_sibling(self, name=None, attrs={}, text=None, **kwargs):
+    def find_next_sibling(self, name=None, attrs=OrderedDict(), text=None, **kwargs):
         """Returns the closest sibling to this Tag that matches the
         given criteria and appears after this Tag in the document."""
         return self._find_one(self.find_next_siblings, name, attrs, text,
                              **kwargs)
     findNextSibling = find_next_sibling  # BS3
 
-    def find_next_siblings(self, name=None, attrs={}, text=None, limit=None,
+    def find_next_siblings(self, name=None, attrs=OrderedDict(), text=None, limit=None,
                            **kwargs):
         """Returns the siblings of this Tag that match the given
         criteria and appear after this Tag in the document."""
@@ -473,14 +462,14 @@ class PageElement(object):
     findNextSiblings = find_next_siblings   # BS3
     fetchNextSiblings = find_next_siblings  # BS2
 
-    def find_previous(self, name=None, attrs={}, text=None, **kwargs):
+    def find_previous(self, name=None, attrs=OrderedDict(), text=None, **kwargs):
         """Returns the first item that matches the given criteria and
         appears before this Tag in the document."""
         return self._find_one(
             self.find_all_previous, name, attrs, text, **kwargs)
     findPrevious = find_previous  # BS3
 
-    def find_all_previous(self, name=None, attrs={}, text=None, limit=None,
+    def find_all_previous(self, name=None, attrs=OrderedDict(), text=None, limit=None,
                         **kwargs):
         """Returns all items that match the given criteria and appear
         before this Tag in the document."""
@@ -489,14 +478,14 @@ class PageElement(object):
     findAllPrevious = find_all_previous  # BS3
     fetchPrevious = find_all_previous    # BS2
 
-    def find_previous_sibling(self, name=None, attrs={}, text=None, **kwargs):
+    def find_previous_sibling(self, name=None, attrs=OrderedDict(), text=None, **kwargs):
         """Returns the closest sibling to this Tag that matches the
         given criteria and appears before this Tag in the document."""
         return self._find_one(self.find_previous_siblings, name, attrs, text,
                              **kwargs)
     findPreviousSibling = find_previous_sibling  # BS3
 
-    def find_previous_siblings(self, name=None, attrs={}, text=None,
+    def find_previous_siblings(self, name=None, attrs=OrderedDict(), text=None,
                                limit=None, **kwargs):
         """Returns the siblings of this Tag that match the given
         criteria and appear before this Tag in the document."""
@@ -505,7 +494,7 @@ class PageElement(object):
     findPreviousSiblings = find_previous_siblings   # BS3
     fetchPreviousSiblings = find_previous_siblings  # BS2
 
-    def find_parent(self, name=None, attrs={}, **kwargs):
+    def find_parent(self, name=None, attrs=OrderedDict(), **kwargs):
         """Returns the closest parent of this Tag that matches the given
         criteria."""
         # NOTE: We can't use _find_one because findParents takes a different
@@ -517,7 +506,7 @@ class PageElement(object):
         return r
     findParent = find_parent  # BS3
 
-    def find_parents(self, name=None, attrs={}, limit=None, **kwargs):
+    def find_parents(self, name=None, attrs=OrderedDict(), limit=None, **kwargs):
         """Returns the parents of this Tag that match the given
         criteria."""
 
@@ -561,7 +550,7 @@ class PageElement(object):
                 result = (element for element in generator
                           if isinstance(element, Tag))
                 return ResultSet(strainer, result)
-            elif isinstance(name, basestring):
+            elif isinstance(name, str):
                 # Optimization to find all tags with a given name.
                 result = (element for element in generator
                           if isinstance(element, Tag)
@@ -712,7 +701,7 @@ class PageElement(object):
         return self.parents
 
 
-class NavigableString(unicode, PageElement):
+class NavigableString(str, PageElement):
 
     PREFIX = ''
     SUFFIX = ''
@@ -725,10 +714,10 @@ class NavigableString(unicode, PageElement):
         passed in to the superclass's __new__ or the superclass won't know
         how to handle non-ASCII characters.
         """
-        if isinstance(value, unicode):
-            u = unicode.__new__(cls, value)
+        if isinstance(value, str):
+            u = str.__new__(cls, value)
         else:
-            u = unicode.__new__(cls, value, DEFAULT_OUTPUT_ENCODING)
+            u = str.__new__(cls, value, DEFAULT_OUTPUT_ENCODING)
         u.setup()
         return u
 
@@ -739,7 +728,7 @@ class NavigableString(unicode, PageElement):
         return type(self)(self)
 
     def __getnewargs__(self):
-        return (unicode(self),)
+        return (str(self),)
 
     def __getattr__(self, attr):
         """text.string gives you text. This is for backwards
@@ -836,15 +825,15 @@ class Tag(PageElement):
         self.namespace = namespace
         self.prefix = prefix
         if attrs is None:
-            attrs = {}
+            attrs = OrderedDict()
         elif attrs:
             if builder is not None and builder.cdata_list_attributes:
                 attrs = builder._replace_cdata_list_attribute_values(
                     self.name, attrs)
             else:
-                attrs = dict(attrs)
+                attrs = OrderedDict(attrs)
         else:
-            attrs = dict(attrs)
+            attrs = OrderedDict(attrs)
         self.attrs = attrs
         self.contents = []
         self.setup(parent, previous)
@@ -1076,26 +1065,11 @@ class Tag(PageElement):
 
     def __repr__(self, encoding="unicode-escape"):
         """Renders this tag as a string."""
-        if PY3K:
-            # "The return value must be a string object", i.e. Unicode
-            return self.decode()
-        else:
-            # "The return value must be a string object", i.e. a bytestring.
-            # By convention, the return value of __repr__ should also be
-            # an ASCII string.
-            return self.encode(encoding)
-
-    def __unicode__(self):
+        # "The return value must be a string object", i.e. Unicode
         return self.decode()
 
     def __str__(self):
-        if PY3K:
-            return self.decode()
-        else:
-            return self.encode()
-
-    if PY3K:
-        __str__ = __repr__ = __unicode__
+        return self.decode()
 
     def encode(self, encoding=DEFAULT_OUTPUT_ENCODING,
                indent_level=None, formatter="minimal",
@@ -1128,7 +1102,7 @@ class Tag(PageElement):
 
         # First off, turn a string formatter into a function. This
         # will stop the lookup from happening over and over again.
-        if not isinstance(formatter, collections.Callable):
+        if not isinstance(formatter, Callable):
             formatter = self._formatter_for_name(formatter)
 
         attrs = []
@@ -1139,8 +1113,8 @@ class Tag(PageElement):
                 else:
                     if isinstance(val, list) or isinstance(val, tuple):
                         val = ' '.join(val)
-                    elif not isinstance(val, basestring):
-                        val = unicode(val)
+                    elif not isinstance(val, str):
+                        val = str(val)
                     elif (
                         isinstance(val, AttributeValueWithCharsetSubstitution)
                         and eventual_encoding is not None):
@@ -1148,7 +1122,7 @@ class Tag(PageElement):
 
                     text = self.format_string(val, formatter)
                     decoded = (
-                        unicode(key) + '='
+                        str(key) + '='
                         + EntitySubstitution.quoted_attribute_value(text))
                 attrs.append(decoded)
         close = ''
@@ -1232,7 +1206,7 @@ class Tag(PageElement):
         """
         # First off, turn a string formatter into a function. This
         # will stop the lookup from happening over and over again.
-        if not isinstance(formatter, collections.Callable):
+        if not isinstance(formatter, Callable):
             formatter = self._formatter_for_name(formatter)
 
         pretty_print = (indent_level is not None)
@@ -1258,7 +1232,7 @@ class Tag(PageElement):
 
         # First off, turn a string formatter into a function. This
         # will stop the lookup from happening over and over again.
-        if not isinstance(formatter, collections.Callable):
+        if not isinstance(formatter, Callable):
             formatter = self._formatter_for_name(formatter)
 
         is_xmlparent = self.name.lower() in EBOOK_XML_PARENT_TAGS
@@ -1336,7 +1310,7 @@ class Tag(PageElement):
         """
         # First off, turn a string formatter into a function. This
         # will stop the lookup from happening over and over again.
-        if not isinstance(formatter, collections.Callable):
+        if not isinstance(formatter, Callable):
             formatter = self._formatter_for_name(formatter)
 
         is_xmlparent = self.name.lower() in EBOOK_XML_PARENT_TAGS
@@ -1359,7 +1333,7 @@ class Tag(PageElement):
     def serialize_xhtml(self, eventual_encoding=DEFAULT_OUTPUT_ENCODING, formatter="minimal"):
         # First off, turn a string formatter into a function. This
         # will stop the lookup from happening over and over again.
-        if not isinstance(formatter, collections.Callable):
+        if not isinstance(formatter, Callable):
             formatter = self._formatter_for_name(formatter)
 
         prefix = ''
@@ -1382,16 +1356,19 @@ class Tag(PageElement):
                     ntext = (str(key) + '=' + EntitySubstitution.quoted_attribute_value(text))
                 attrs.append(ntext)
 
+        contents = self.serialize_xhtml_contents(eventual_encoding, formatter)
+
+        in_xml_ns = self.namespace != 'http://www.w3.org/1999/xhtml'
+        testcontents = contents.strip()
+
         if self.prefix:
             prefix = self.prefix + ":"
 
-        if self.name in VOID_TAGS:
+        if self.name in VOID_TAGS or (in_xml_ns and testcontents==""):
             close = '/'
         else:
             closeTag = '</%s%s>' % (prefix, self.name)
 
-        contents = self.serialize_xhtml_contents(eventual_encoding, formatter)
-
         # strip extraneous whitespace before the primary closing tag
         if self.name in SPECIAL_HANDLING_TAGS:
             contents = contents.strip()
@@ -1419,7 +1396,7 @@ class Tag(PageElement):
 
         # First off, turn a string formatter into a function. This
         # will stop the lookup from happening over and over again.
-        if not isinstance(formatter, collections.Callable):
+        if not isinstance(formatter, Callable):
             formatter = self._formatter_for_name(formatter)
 
         s = []
@@ -1443,7 +1420,7 @@ class Tag(PageElement):
 
         # First off, turn a string formatter into a function. This
         # will stop the lookup from happening over and over again.
-        if not isinstance(formatter, collections.Callable):
+        if not isinstance(formatter, Callable):
             formatter = self._formatter_for_name(formatter)
 
         is_structural = self.name in STRUCTURAL_TAGS
@@ -1473,14 +1450,10 @@ class Tag(PageElement):
                 attribs.append(decoded)
             atts = " " + " ".join(attribs)
 
-        prefix = ''
-        if self.prefix:
-            prefix = self.prefix + ":"
-
-        is_void_tag = self.name in VOID_TAGS
 
         # get tag content
         contents=""
+        is_void_tag = self.name in VOID_TAGS
         if not is_void_tag:
             if is_structural:
                 contents = self.prettyprint_xhtml_contents(indent_level+1, eventual_encoding, formatter, indent_chars)
@@ -1491,15 +1464,18 @@ class Tag(PageElement):
             # This is the 'document root' object.
             return contents
 
+        in_xml_ns = self.namespace != 'http://www.w3.org/1999/xhtml'
+        testcontents = contents.strip()
+        single = self.name in VOID_TAGS or (in_xml_ns and testcontents == "")
+
+        prefix = ''
+        if self.prefix:
+            prefix = self.prefix + ":"
+
         is_keepwhitespace = self.name in PRESERVE_WHITESPACE_TAGS
         if not is_keepwhitespace and not is_inline:
             contents = contents.rstrip()
 
-        single = is_void_tag
-        # for xhtml serialization with self-closing non-void tags
-        # uncomment the following line
-        # single = single or (contents == "")
-
         indent_space = (indent_chars * (indent_level - 1))
 
         # handle self-closed tags with no content first
@@ -1539,7 +1515,7 @@ class Tag(PageElement):
         """
         # First off, turn a string formatter into a function. This
         # will stop the lookup from happening over and over again.
-        if not isinstance(formatter, collections.Callable):
+        if not isinstance(formatter, Callable):
             formatter = self._formatter_for_name(formatter)
 
         is_structural = self.name in STRUCTURAL_TAGS
@@ -1645,7 +1621,7 @@ class Tag(PageElement):
 
     #Soup methods
 
-    def find(self, name=None, attrs={}, recursive=True, text=None,
+    def find(self, name=None, attrs=OrderedDict(), recursive=True, text=None,
              **kwargs):
         """Return only the first child of this Tag matching the given
         criteria."""
@@ -1656,7 +1632,7 @@ class Tag(PageElement):
         return r
     findChild = find
 
-    def find_all(self, name=None, attrs={}, recursive=True, text=None,
+    def find_all(self, name=None, attrs=OrderedDict(), recursive=True, text=None,
                  limit=None, **kwargs):
         """Extracts a list of Tag objects that match the given
         criteria.  You can specify the name of the Tag and any
@@ -1939,7 +1915,7 @@ class SoupStrainer(object):
     """Encapsulates a number of ways of matching a markup element (tag or
     text)."""
 
-    def __init__(self, name=None, attrs={}, text=None, **kwargs):
+    def __init__(self, name=None, attrs=OrderedDict(), text=None, **kwargs):
         self.name = self._normalize_search_value(name)
         if not isinstance(attrs, dict):
             # Treat a non-dict value for attrs as a search for the 'class'
@@ -1959,7 +1935,7 @@ class SoupStrainer(object):
                 attrs.update(kwargs)
             else:
                 attrs = kwargs
-        normalized_attrs = {}
+        normalized_attrs = OrderedDict()
         for key, value in list(attrs.items()):
             normalized_attrs[key] = self._normalize_search_value(value)
 
@@ -1969,7 +1945,7 @@ class SoupStrainer(object):
     def _normalize_search_value(self, value):
         # Leave it alone if it's a Unicode string, a callable, a
         # regular expression, a boolean, or None.
-        if (isinstance(value, text_type) or isinstance(value, collections.Callable) or hasattr(value, 'match')
+        if (isinstance(value, str) or isinstance(value, Callable) or hasattr(value, 'match')
             or isinstance(value, bool) or value is None):
             return value
 
@@ -1982,7 +1958,7 @@ class SoupStrainer(object):
             new_value = []
             for v in value:
                 if (hasattr(v, '__iter__') and not isinstance(v, bytes)
-                    and not isinstance(v, text_type)):
+                    and not isinstance(v, str)):
                     # This is almost certainly the user's mistake. In the
                     # interests of avoiding infinite loops, we'll let
                     # it through as-is rather than doing a recursive call.
@@ -1992,9 +1968,7 @@ class SoupStrainer(object):
             return new_value
 
         # Otherwise, convert it into a Unicode string.
-        # The unicode(str()) thing is so this will do the same thing on Python 2
-        # and Python 3.
-        return unicode(unicode(value))
+        return str(value)
 
     def __str__(self):
         if self.text:
@@ -2002,14 +1976,14 @@ class SoupStrainer(object):
         else:
             return "%s|%s" % (self.name, self.attrs)
 
-    def search_tag(self, markup_name=None, markup_attrs={}):
+    def search_tag(self, markup_name=None, markup_attrs=OrderedDict()):
         found = None
         markup = None
         if isinstance(markup_name, Tag):
             markup = markup_name
             markup_attrs = markup
         call_function_with_tag_data = (
-            isinstance(self.name, collections.Callable)
+            isinstance(self.name, Callable)
             and not isinstance(markup_name, Tag))
 
         if ((not self.name)
@@ -2026,7 +2000,7 @@ class SoupStrainer(object):
                         if hasattr(markup_attrs, 'get'):
                             markup_attr_map = markup_attrs
                         else:
-                            markup_attr_map = {}
+                            markup_attr_map = OrderedDict()
                             for k, v in markup_attrs:
                                 markup_attr_map[k] = v
                     attr_value = markup_attr_map.get(attr)
@@ -2048,7 +2022,7 @@ class SoupStrainer(object):
         found = None
         # If given a list of items, scan it for a text element that
         # matches.
-        if hasattr(markup, '__iter__') and not isinstance(markup, (Tag, basestring)):
+        if hasattr(markup, '__iter__') and not isinstance(markup, (Tag, str)):
             for element in markup:
                 if isinstance(element, NavigableString) \
                        and self.search(element):
@@ -2061,7 +2035,7 @@ class SoupStrainer(object):
                 found = self.search_tag(markup)
         # If it's text, make sure the text matches.
         elif isinstance(markup, NavigableString) or \
-                 isinstance(markup, basestring):
+                 isinstance(markup, str):
             if not self.name and not self.attrs and self._matches(markup, self.text):
                 found = markup
         else:
@@ -2075,7 +2049,7 @@ class SoupStrainer(object):
         if isinstance(markup, list) or isinstance(markup, tuple):
             # This should only happen when searching a multi-valued attribute
             # like 'class'.
-            if (isinstance(match_against, unicode)
+            if (isinstance(match_against, str)
                 and ' ' in match_against):
                 # A bit of a special case. If they try to match "foo
                 # bar" on a multivalue attribute's value, only accept
@@ -2095,7 +2069,7 @@ class SoupStrainer(object):
             # True matches any non-None value.
             return markup is not None
 
-        if isinstance(match_against, collections.Callable):
+        if isinstance(match_against, Callable):
             return match_against(markup)
 
         # Custom callables take the tag as an argument, but all
@@ -2110,7 +2084,7 @@ class SoupStrainer(object):
             # None matches None, False, an empty string, an empty list, and so on.
             return not match_against
 
-        if isinstance(match_against, unicode):
+        if isinstance(match_against, str):
             # Exact string match
             return markup == match_against
 
